#
# Copyright (c) Daniel Sheffield 2023
# All rights reserved
#
# THIS SOFTWARE IS PROVIDED AS IS WITHOUT WARRANTY
from itertools import chain
from time import time
from typing import Iterable
from io import StringIO
import os
from urllib.parse import urlencode
from bottle import (
    route,
    request,
    run,
    response,
    abort,
    FormsDict,
    redirect,
    template,
    HTTPError,
    static_file,
)
from psycopg import connect
from psycopg.sql import SQL, Literal
import matplotlib.pyplot as plt
import seaborn as sns
from multiprocessing import Lock
from ..activities.Plot import (
    get_data,
)
from ..data.QueryManager import QueryManager, display_mapper
from ..data.util import(
    get_include_exclude,
    get_where_include_exclude
)
import matplotlib
matplotlib.use('agg')

def line(pivot, ylabel=None, xlabel=None):
    ax = sns.lineplot(data=pivot, markers=True)
    ax.set_xlabel(xlabel, color='#cccccc')
    ax.set_ylabel(ylabel, color='#cccccc')
    ax.axes.tick_params(colors='#cccccc', which='both')
    for _, spine in ax.spines.items():
        spine.set_color('#cccccc')

ALL_UNITS = {'g', 'kg', 'mL', 'L', 'Pieces', 'Bunches', 'Bags'}
PARAMS = { 'group', 'category', 'product', 'unit', 'tag' }
CACHE = dict()

host = f"host={os.getenv('HOST')}"
db = f"dbname={os.getenv('DB', 'grocery')}"
user = f"user={os.getenv('USER', 'das')}"
password = f"password={os.getenv('PASSWORD','')}"
if not password.split('=',1)[1]:
    password = ''
conn = connect(f"{host} {db} {user} {password}")
sns.set_theme(style='darkgrid', palette='pastel')

def get_product_rollup_statement(filters, having=None):
    where = [ get_where_include_exclude(
        k[0], "name", include, exclude
    ) for k, (include, exclude) in filters.items() ]
    return SQL('\n').join([
        SQL("""
SELECT
  count(DISTINCT p.id) AS "Products",
  count(DISTINCT c.id) AS "Categories",
  count(DISTINCT g.id) AS "Groups",
  p.name AS "Product",
  c.name AS "Category",
  g.name AS "Group"
FROM products p
JOIN categories c ON p.category_id = c.id
JOIN groups g ON c.group_id = g.id
"""),
        SQL("""
WHERE {where}
""").format(where=SQL("\nAND").join(where)) if where else SQL(''),
        SQL("""
GROUP BY ROLLUP (g.name, c.name, p.name)
"""),
        SQL("""
HAVING {having}
ORDER BY "Product", "Category", "Group"
""").format(having=having) if having else SQL('')
    ])

def get_filter(query: FormsDict, allow: Iterable[str] = None):
    return {
        k: list(map(set, get_include_exclude(
            (query[k] or 'kg' if k == 'unit' else query.getall(k)) if k in query else None
        ))) for k in sorted(set(query.keys()) | set(allow)) if allow is None or k in allow
    }

def get_query_param(include, exclude):
    return '!'.join([
        '|'.join(sorted(include)),
        *([ '|'.join(sorted(exclude)) ] if exclude else [ ]),
    ])

def get_query(**params):
    return '&'.join([
        urlencode([ (k, get_query_param(*params[k])) for k in sorted(params) ])
    ])

def normalize_query(query: FormsDict, allow: Iterable[str] = None):
    return get_query(**get_filter(query, allow=allow))

def get_form(action, method, filter_data, data):
    keys = sorted(filter(lambda x: x not in ('unit', 'tag'), filter_data), key=lambda x: {
        'product': 0,
        'category': 1,
        'group': 2,
    }[x])
    in_chart = data['$/unit'].apply(lambda x: (x or False) and True)
    group = data[in_chart].groupby([
        k for k in keys
    ]).size().reset_index()[[
        k for k in keys
    ]]
    return template(
        'app/rest/form',
        action=action,
        method=method,
        header=[
            template(
                'app/rest/filter-heading',
                fname=k,
                first=(idx == 0),
            ) for idx, k in enumerate(keys)
        ],
        **{
            k: {
                "name": k,
                "_include": {
                    "options": sorted(map(lambda x: {
                        "selected": x[0],
                        "value": x[1],
                    }, chain(
                        map(lambda x: (True, x), filter_data[k][0]),
                        map(lambda x: (False, x), set([
                            x[k] for _, x in group.iterrows()
                        ]) - filter_data[k][0])
                    )), key=lambda x: x["display"] if "display" in x else x["value"])
                },
                "_exclude": {
                    "options": sorted(map(lambda x: {
                        "selected": x[0],
                        "value": f"!{x[1]}",
                        "display": x[1]
                    }, chain(
                        map(lambda x: (True, x), filter_data[k][1]),
                        map(lambda x: (False, x), set([
                            x[k] for _, x in group.iterrows()
                        ]) - filter_data[k][1])
                    )), key=lambda x: x["display"] if "display" in x else x["value"])
                }
            } for k in keys
        },
        tags={
            "name": "tags",
            "_include": {
                "options": sorted(map(lambda x: {
                    "selected": x[0],
                    "value": x[1],
                },chain(
                    map(lambda x: (True, x), filter_data['tag'][0]),
                    map(lambda x: (False, x), set(chain(*data[in_chart]['tags'])) - filter_data['tag'][0])
                )), key=lambda x: x["display"] if "display" in x else x["value"])
            }
        },
        units={
            "name": "units",
            "options": sorted(map(lambda x: {
                "selected": x[0],
                "value": x[1],
            },chain(
                map(lambda x: (True, x), filter_data['unit'][0]),
                map(lambda x: (False, x), ALL_UNITS - filter_data['unit'][0])
            )), key=lambda x: x["display"] if "display" in x else x["value"])
        }
    )

@route('/grocery/static/<filename:path>')
def send_static(filename):
    return static_file(filename, root='app/rest/static')



global LOCK
LOCK = Lock()

@route('/grocery/trend')
def trend():
    _, _, path, *_ = request.urlparts
    normalized = normalize_query(request.query, allow=PARAMS)

    if request.query_string != normalized:
        return redirect(f'{path}?{normalized}')
    
    loading = template("app/rest/loading", progress=[])
    if request.query_string in CACHE:
        if LOCK.acquire(block=False):
            try:
                return CACHE[request.query_string]["state"]
            finally:
                try:
                    CACHE[request.query_string]["state"] = next(CACHE[request.query_string]["iter"])
                except StopIteration:
                    del CACHE[request.query_string]
                finally:
                    LOCK.release()
        else:
            return CACHE[request.query_string]["state"]

    if LOCK.acquire(block=False):
        if request.query_string in CACHE:
            LOCK.release()
            return CACHE[request.query_string]["state"]
        try:
            CACHE[request.query_string] = {
                "iter": trend_internal(path, request.query),
                "state": loading,
            }
        finally:
            LOCK.release()
    return loading

def trend_internal(path, query):
    progress = []
    try:
        with conn.cursor() as cur:
            query_manager = QueryManager(cur, display_mapper)
            fields = { k: query[k] or None for k in query.keys() if k in PARAMS }
            unit = fields['unit'] = fields['unit'] or 'kg' if 'unit' in fields else 'kg'
            if unit and unit not in ALL_UNITS:
                raise abort(400, f"Unsupported unit {unit}")

            progress.append({ "name": "Loading data", "status": ""})
            yield template("app/rest/loading", progress=progress)
            data = get_data(query_manager, **fields)

            progress[-1]["status"] = "done"
            yield template("app/rest/loading", progress=progress)

            if data.empty:
                raise abort(404, f"No data for {fields}")
            
            progress.append({ "name": "Loading chart", "status": ""})
            yield template("app/rest/loading", progress=progress)
            pivot = data.pivot_table(index=['ts_raw',], columns=['product',], values=['$/unit'], aggfunc='mean')
            pivot.columns = pivot.columns.droplevel()
            plt.figure(facecolor='black', figsize=[16, 9])
            line(pivot, xlabel='Time', ylabel=f'$ / {unit}')
            
            progress[-1]["status"] = "done"
            yield template("app/rest/loading", progress=progress)
            
            progress.append({ "name": "Rendering chart", "status": ""})
            yield template("app/rest/loading", progress=progress)

            f = StringIO()
            plt.savefig(f, format='svg')
            form = get_form(path, 'get', get_filter(request.query, allow=PARAMS), data)
            
            progress[-1]["status"] = "done"
            yield template("app/rest/loading", progress=progress)
            
            final = template("app/rest/trend", form=form, svg=f.getvalue())
            resp = lambda: template("app/rest/trend", form=form, svg=f.getvalue())

    except HTTPError as e:
        resp = lambda exception=e: exception

    finally:
        conn.commit()
     
    yield from iter(resp, lambda started=time(): time() - started > 600)

@route('/grocery/xslt')
def table():
    title = request.query['title'] if 'title' in request.query.keys() else None
    foot = request.query['foot'] == 'True' and "last()" if 'foot' in request.query.keys() else "-1"
    response.content_type = 'application/xhtml+xml; charset=utf-8'
    return template("app/rest/query-to-xml-xslt", title=title, foot=foot)

@route('/grocery/groups')
def groups():
    filters = get_filter(request.query, allow=('group', 'category', 'product'))
    try:
        with conn.cursor() as cur:
            inner = get_product_rollup_statement(
                filters,
                having=SQL("c.name IS NULL")
            )
            xml = cur.execute(SQL("""
SELECT query_to_xml_and_xmlschema({q}, false, false, ''::text)
""").format(q=Literal(SQL("""SELECT
    "Products",
    "Categories",
    COALESCE("Group", "Groups"||'') "Group"
FROM (
{inner}
) q""").format(inner=inner).as_string(cur)))).fetchone()[0]
    finally:
        conn.commit()
    response.content_type = 'application/xhtml+xml; charset=utf-8'
    return template(
        "app/rest/query-to-xml",
        title="Groups",
        foot=True,
        xml=xml
    )

@route('/grocery/categories')
def categories():
    filters = get_filter(request.query, allow=('group', 'category', 'product'))
    try:
        with conn.cursor() as cur:
            inner = get_product_rollup_statement(
                filters,
                having=SQL("p.name IS NULL AND (c.name IS NOT NULL OR g.name IS NULL)")
            )
            xml = cur.execute(SQL("""
SELECT query_to_xml_and_xmlschema({q}, false, false, ''::text)
""").format(q=Literal(SQL("""SELECT
    "Products",
    COALESCE("Category", "Categories"||'') "Category",
    COALESCE("Group", "Groups"||'') "Group"
FROM (
{inner}
) q""").format(inner=inner).as_string(cur)))).fetchone()[0]
    finally:
        conn.commit()
    response.content_type = 'application/xhtml+xml; charset=utf-8'
    return template(
        "app/rest/query-to-xml",
        title="Categories",
        foot=True,
        xml=xml
    )

@route('/grocery/products')
def products():
    filters = get_filter(request.query, allow=('group', 'category', 'product'))
    try:
        with conn.cursor() as cur:
            inner = get_product_rollup_statement(
                filters,
                having=SQL("p.name IS NOT NULL OR g.name IS NULL")
            )
            xml = cur.execute(SQL("""
SELECT query_to_xml_and_xmlschema({q}, false, false, ''::text)
""").format(q=Literal(SQL("""SELECT
    --"Transactions",
    COALESCE("Product", "Products"||'') "Product",
    COALESCE("Category", "Categories"||'') "Category",
    COALESCE("Group", "Groups"||'') "Group"
FROM (
{inner}
) q""").format(inner=inner).as_string(cur)))).fetchone()[0]
    finally:
        conn.commit()
    response.content_type = 'application/xhtml+xml; charset=utf-8'
    return template(
        "app/rest/query-to-xml",
        title="Products",
        foot=True,
        xml=xml
    )

@route('/grocery/tags')
def tags():
    try:
        with conn.cursor() as cur:
            inner = SQL('\n').join([SQL("""
SELECT count(DISTINCT txn.id) AS "Uses", tg.name AS "Name"
FROM tags tg
JOIN tags_map tm ON tg.id = tm.tag_id
JOIN transactions txn ON txn.id = tm.transaction_id
GROUP BY tg.name
ORDER BY 1 DESC, 2
""")]).as_string(cur)
            xml = cur.execute(SQL("""
SELECT query_to_xml_and_xmlschema({inner}, false, false, ''::text)
""").format(inner=Literal(inner))).fetchone()[0]
    finally:
        conn.commit()
    response.content_type = 'application/xhtml+xml; charset=utf-8'
    return template(
        "app/rest/query-to-xml",
        title="Tags",
        foot="false",
        xml=xml
    )

run(host='0.0.0.0', port=6772, server='gunicorn')
